package net.keithlantz.threadedserver;

import android.app.Activity;
import android.os.Bundle;
import android.os.Handler;
import android.net.NetworkInfo;
import android.net.ConnectivityManager;
import android.content.Context;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.Button;
import android.widget.ScrollView;
import android.widget.TextView;
import android.widget.Toast;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.spi.SelectorProvider;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.io.IOException;
import java.lang.String;

public class ThreadedServer extends Activity {

	private String ipAddress;
	private ServerSocket serverSocket;
	private ServerSocketChannel serverSocketChannel;
	private static final int PORT = 31236;

	private Button stopThreads;
	private Button startThread;
	volatile boolean runThreads;

	private ScrollView scroll;
	private TextView log;

	private Handler handler = new Handler();

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		setContentView(R.layout.main);
		log = (TextView)findViewById(R.id.log);
		scroll = (ScrollView)findViewById(R.id.scroll);

		runThreads = false;

		stopThreads = (Button)findViewById(R.id.stopThreads);
		stopThreads.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				runThreads = false;
			}
		});

		startThread = (Button)findViewById(R.id.startThread);
		startThread.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				if (!runThreads) {
					boolean networkAvailable = isNetworkAvailable();
					log.append(networkAvailable ? "Network available.\n" : "Network unavailable.\n");
					scroll.fullScroll(View.FOCUS_DOWN);
					if (networkAvailable) {
						InetAddress inetAddress = getInetAddress();
						ipAddress = inetAddress != null ? inetAddress.getHostAddress().toString() : "";
						runThreads = true;
						new Thread(new ServerThread()).start();
					}
				}
			}
		});
	}

	@Override
	public void onDestroy() {
		super.onDestroy();
		runThreads = false;
	}

	private boolean isNetworkAvailable() {
		NetworkInfo networkInfo = ((ConnectivityManager)getSystemService(Context.CONNECTIVITY_SERVICE)).getActiveNetworkInfo();
		return networkInfo != null && networkInfo.isConnected();
	}

	private InetAddress getInetAddress() {
		try {
			Enumeration<NetworkInterface> enum_networkInterface = NetworkInterface.getNetworkInterfaces();
			while (enum_networkInterface.hasMoreElements()) {
				NetworkInterface networkInterface = enum_networkInterface.nextElement();
				Enumeration<InetAddress> enum_inetAddress = networkInterface.getInetAddresses();
				while (enum_inetAddress.hasMoreElements()) {
					InetAddress inetAddress = enum_inetAddress.nextElement();
					if (!inetAddress.isLoopbackAddress()) return inetAddress;
				}
			}
		} catch (SocketException e) {
			log.append(e.toString()+"\n");
			scroll.fullScroll(View.FOCUS_DOWN);
		}
		return null;
	}

	public class AcceptThread implements Runnable {
		private SocketChannel sc;
		private String socketAddressStr;
		private boolean runChild;
		public AcceptThread(SocketChannel socketChannel) {
			sc = socketChannel;
			socketAddressStr = "IP "+sc.socket().getInetAddress().getHostAddress().toString()+" Port "+sc.socket().getPort();
			runChild = true;
		}

		public void run() {
			handler.post(new Runnable() {
				@Override
				public void run() {
					log.append("Child thread started (Client "+socketAddressStr+").\n");
					scroll.fullScroll(View.FOCUS_DOWN);
				}
			});
			try {
				sc.configureBlocking(false);
				while (runThreads && runChild) {
					ByteBuffer buffer = ByteBuffer.allocateDirect(1024);
					buffer.clear();
					int bytesRead = sc.read(buffer);
					if (bytesRead == -1) {
						runChild = false;
						sc.close();
					} else if (bytesRead > 0) {
						buffer.flip();
						Charset charset = Charset.forName("UTF-8");
						CharsetDecoder decoder = charset.newDecoder();
						final String buf = decoder.decode(buffer).toString();
						handler.post(new Runnable() {
							@Override
							public void run() {
								log.append("(Client IP "+socketAddressStr+") "+buf);
								scroll.fullScroll(View.FOCUS_DOWN);
							}
						});
					}
				}
			} catch(IOException e) {
				final String err = e.toString();
				handler.post(new Runnable() {
					@Override
					public void run() {
						log.append(err+"\n");
						scroll.fullScroll(View.FOCUS_DOWN);
					}
				});
			}
			try {
				sc.close();
			} catch(IOException e) {
				final String err = e.toString();
				handler.post(new Runnable() {
					@Override
					public void run() {
						log.append(err+"\n");
						scroll.fullScroll(View.FOCUS_DOWN);
					}
				});
			}
			handler.post(new Runnable() {
				@Override
				public void run() {
					log.append("Child thread terminated (Client "+socketAddressStr+").\n");
					scroll.fullScroll(View.FOCUS_DOWN);
				}
			});
		}
	}

	public class ServerThread implements Runnable {
		public void run() {
			handler.post(new Runnable() {
				@Override
				public void run() {
					log.append("Main thread started (Server IP "+ipAddress+" Port "+PORT+").\n");
					scroll.fullScroll(View.FOCUS_DOWN);
				}
			});
			try {
				serverSocketChannel = ServerSocketChannel.open();
				serverSocketChannel.configureBlocking(false);
				serverSocket = serverSocketChannel.socket();
				serverSocket.bind(new InetSocketAddress(PORT));
				while(runThreads) {
					SocketChannel socketChannel = serverSocketChannel.accept();
					if (socketChannel != null)
						new Thread(new AcceptThread(socketChannel)).start();
				}
			} catch(IOException e) {
				final String err = e.toString();
				handler.post(new Runnable() {
					@Override
					public void run() {
						log.append(err+"\n");
						scroll.fullScroll(View.FOCUS_DOWN);
					}
				});
			}
			try {
				serverSocketChannel.close();
			} catch(IOException e) {
				final String err = e.toString();
				handler.post(new Runnable() {
					@Override
					public void run() {
						log.append(err+"\n");
						scroll.fullScroll(View.FOCUS_DOWN);
					}
				});
			}
			handler.post(new Runnable() {
				@Override
				public void run() {
					log.append("Main thread terminated (Server "+ipAddress+" Port "+PORT+").\n");
					scroll.fullScroll(View.FOCUS_DOWN);
				}
			});
		}
	}
}